"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
exports.__esModule = true;
exports.importAliasRule = void 0;
var alias_config_1 = require("#src/alias-config");
var fs_extra_1 = require("fs-extra");
var path_1 = require("path");
var slash_1 = __importDefault(require("slash"));
function isPermittedRelativeImport(importModuleName, relativeImportOverrides, filepath, projectBaseDir) {
    var importParts = importModuleName.split("/");
    var relativeDepth = importParts.filter(function (moduleNamePart) { return moduleNamePart === ".."; }).length;
    var configs = __spreadArray([], relativeImportOverrides, true);
    configs.sort(function (a, b) { return b.depth - a.depth; }); // rank depth descending
    for (var _i = 0, configs_1 = configs; _i < configs_1.length; _i++) {
        var config = configs_1[_i];
        if (("path" in config && filepath.includes((0, path_1.resolve)(config.path))) ||
            ("pattern" in config &&
                new RegExp(config.pattern).test((0, path_1.relative)(projectBaseDir, filepath)))) {
            return relativeDepth <= config.depth;
        }
    }
    return false;
}
function getAliasSuggestion(importModuleName, aliasConfigs, absoluteDir) {
    var currentAliasConfig = aliasConfigs.find(function (_a) {
        var alias = _a.alias;
        var baseModulePath = importModuleName.split("/")[0];
        return baseModulePath === alias;
    });
    var absoluteModulePath = undefined;
    if (importModuleName.trim().charAt(0) === ".") {
        absoluteModulePath = (0, path_1.join)(absoluteDir, importModuleName);
    }
    else if (currentAliasConfig) {
        absoluteModulePath = importModuleName
            .replace(currentAliasConfig.alias, currentAliasConfig.path.absolute)
            .replace(/\//g, path_1.sep);
    }
    if (absoluteModulePath) {
        var bestAliasConfig = getBestAliasConfig(aliasConfigs, currentAliasConfig, absoluteModulePath);
        if (bestAliasConfig && bestAliasConfig !== currentAliasConfig) {
            return (0, slash_1["default"])(absoluteModulePath.replace(bestAliasConfig.path.absolute, bestAliasConfig.alias));
        }
    }
}
function getBestAliasConfig(aliasConfigs, currentAlias, absoluteModulePath) {
    var importPathParts = absoluteModulePath.split(path_1.sep);
    return aliasConfigs.reduce(function (currentBest, potentialAlias) {
        var aliasPathParts = potentialAlias.path.absolute.split(path_1.sep);
        var isValidAlias = aliasPathParts.reduce(function (isValid, aliasPathPart, index) {
            return isValid && importPathParts[index] === aliasPathPart;
        }, true);
        var isMoreSpecificAlias = !currentBest ||
            potentialAlias.path.absolute.length >
                currentBest.path.absolute.length;
        return isValidAlias && isMoreSpecificAlias
            ? potentialAlias
            : currentBest;
    }, currentAlias);
}
/**
 * This should match the type definition of ImportAliasOptions
 */
var schemaProperties = {
    aliasConfigPath: {
        description: "Alternative path to look for a TSConfig/JSConfig",
        type: "string"
    },
    aliasImportFunctions: {
        type: "array",
        description: "Function names which supports aliases as its first function." +
            " Examples are `require` and `mock` for `jest.mock`." +
            " Please submit an issue if position parameters needs support",
        items: {
            type: "string"
        },
        "default": ["require", "mock"]
    },
    relativeImportOverrides: {
        type: "array",
        "default": [],
        items: {
            anyOf: [
                {
                    type: "object",
                    properties: {
                        path: {
                            type: "string",
                            description: "The starting path from which a relative depth is accepted." +
                                " Required if `pattern` is not provided"
                        },
                        depth: {
                            type: "number",
                            description: "A positive number which represents the" +
                                " relative depth that is acceptable for the associated path."
                        }
                    }
                },
                {
                    type: "object",
                    properties: {
                        pattern: {
                            type: "string",
                            description: "The pattern to match against the filename for from" +
                                " which a relative depth is accepted." +
                                " Required if `path` is not provided"
                        },
                        depth: {
                            type: "number",
                            description: "A positive number which represents the" +
                                " relative depth that is acceptable for the associated path."
                        }
                    }
                },
            ]
        }
    }
};
var importAliasRule = {
    meta: {
        docs: {
            description: "Suggests shortest matching alias in TSConfig/JSConfig",
            category: "Suggestions",
            recommended: true,
            url: "https://github.com/limegrass/eslint-plugin-import-alias/blob/HEAD/docs/rules/import-alias.md",
            suggestion: true
        },
        fixable: "code",
        schema: [
            {
                type: "object",
                properties: schemaProperties
            },
        ],
        type: "suggestion"
    },
    create: function (context) {
        var _a;
        var reportProgramError = function (errorMessage) {
            return {
                Program: function (node) {
                    context.report({ node: node, message: errorMessage });
                }
            };
        };
        /**
         * cwd seems to resolve to the path where the .eslintrc.js being used is found.
         * Thus, it is the appropriate place to append the aliasConfigPath from since that is
         * where the user would specify their aliasConfigPath relative to.
         * We can also otherwise resolve the tsconfig/jsconfig from the dirname(filepath),
         * which tsconfig-paths will attempt automatically for `tsconfig.json` and `jsconfig.json`
         */
        var cwd = context.getCwd();
        var filepath = (0, path_1.resolve)(context.getFilename());
        var absoluteDir = (0, path_1.dirname)(filepath);
        if (!(0, fs_extra_1.existsSync)(absoluteDir)) {
            return reportProgramError("a filepath must be provided, try with --stdin-filename, " +
                "call eslint on a file, " +
                "or save your buffer as a file and restart eslint in your editor.");
        }
        var _b = context.options[0] || {}, aliasConfigPath = _b.aliasConfigPath, _c = _b.aliasImportFunctions, aliasImportFunctions = _c === void 0 ? schemaProperties.aliasImportFunctions["default"] : _c, _d = _b.relativeImportOverrides, relativeImportOverrides = _d === void 0 ? [] : _d; // No idea what the other array values are
        var projectBaseDir;
        var aliasesResult;
        try {
            var tsconfig = (0, alias_config_1.loadTsconfig)(cwd, aliasConfigPath, filepath);
            var configDir = (0, path_1.dirname)(tsconfig.configFileAbsolutePath);
            projectBaseDir = (0, path_1.join)(configDir, (_a = tsconfig.baseUrl) !== null && _a !== void 0 ? _a : "");
            aliasesResult = (0, alias_config_1.loadAliasConfigs)(tsconfig, projectBaseDir);
        }
        catch (error) {
            if (error instanceof Error) {
                return reportProgramError(error.message);
            }
            throw error;
        }
        var getReportDescriptor = function (_a, importModuleName) {
            var moduleStart = _a[0], moduleEnd = _a[1];
            // preserve user quote style
            var quotelessRange = [moduleStart + 1, moduleEnd - 1];
            if (isPermittedRelativeImport(importModuleName, relativeImportOverrides, filepath, projectBaseDir)) {
                return undefined;
            }
            var aliasSuggestion = getAliasSuggestion(importModuleName, aliasesResult, absoluteDir);
            return aliasSuggestion
                ? {
                    message: "import ".concat(importModuleName, " can be written as ").concat(aliasSuggestion),
                    fix: function (fixer) {
                        return fixer.replaceTextRange(quotelessRange, aliasSuggestion);
                    }
                }
                : undefined;
        };
        var aliasModuleDeclarations = function (node) {
            var _a;
            if (((_a = node.source) === null || _a === void 0 ? void 0 : _a.range) && typeof node.source.value === "string") {
                var suggestion = getReportDescriptor(node.source.range, node.source.value);
                if (suggestion) {
                    context.report(__assign({ node: node }, suggestion));
                }
            }
        };
        return {
            CallExpression: function (node) {
                var identifierNode = node.callee.type === "MemberExpression"
                    ? node.callee.property
                    : node.callee;
                if (identifierNode.type === "Identifier" &&
                    aliasImportFunctions.includes(identifierNode.name) &&
                    node.arguments.length) {
                    var importNameNode = node.arguments[0];
                    if (importNameNode.range &&
                        "value" in importNameNode &&
                        typeof importNameNode.value === "string") {
                        var suggestion = getReportDescriptor(importNameNode.range, importNameNode.value);
                        if (suggestion) {
                            context.report(__assign({ node: node }, suggestion));
                        }
                    }
                }
            },
            ExportAllDeclaration: aliasModuleDeclarations,
            ExportNamedDeclaration: aliasModuleDeclarations,
            ImportDeclaration: aliasModuleDeclarations
        };
    }
};
exports.importAliasRule = importAliasRule;
